# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MFG utilities."""

import pickle

from open_spiel.python.utils import gfile
from open_spiel.python.mfg import distribution


def save_parametric_distribution(dist: distribution.ParametricDistribution,
                                 filename: str):
  """Saves the parametric distribution to a Pickle file."""
  with gfile.Open(filename, "wb") as f:
    pickle.dump(dist.get_params(), f, protocol=pickle.DEFAULT_PROTOCOL)
