// Copyright 2019 DeepMind Technologies Limited
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "open_spiel/games/markov_soccer/markov_soccer.h"

#include "open_spiel/spiel_utils.h"
#include "open_spiel/tests/basic_tests.h"

namespace open_spiel {
namespace markov_soccer {
namespace {

namespace testing = open_spiel::testing;

void BasicMarkovSoccerTests() {
  testing::LoadGameTest("markov_soccer");
  testing::ChanceOutcomesTest(*LoadGame("markov_soccer"));

  constexpr const char big_grid[] =
      "......................\n"
      ".A....................\n"
      "......................\n"
      "......................\n"
      "...........O..........\n"
      "...........O..........\n"
      "...........O..........\n"
      "......................\n"
      "......................\n"
      "....................B.\n"
      "......................\n";

  testing::RandomSimTest(
      *LoadGame("markov_soccer",
                {{"horizon", GameParameter(100)},
                 {"grid", GameParameter(std::string(big_grid))}}),
      100);
}

}  // namespace
}  // namespace markov_soccer
}  // namespace open_spiel

int main(int argc, char **argv) {
  open_spiel::markov_soccer::BasicMarkovSoccerTests();
}
